#include <fstream>
#include <cassert>
#include <cmath>

#include "BPLAKernel.h"

using namespace std;
using namespace boost;
namespace po = boost::program_options;

const float default_score_table[N_RNA][N_RNA] = {
  /* A         C          G          U       */
  { 5.846613, -1.860000, -1.460000, -1.390000}, // A
  {-1.860000,  4.786613, -2.480000, -1.050000}, // C
  {-1.460000, -2.480000,  4.656613, -1.740000}, // G
  {-1.390000, -1.050000, -1.740000,  5.276613}, // U
};

BPLAKernel::Options::
Options(BPLAKernel::Mode mode) 
{
  if (mode==DATABASE) {
    noBP = false;
    gap = -27;
    ext = -0.1;
    alpha = 1.0;
    beta = 0.1;
  }
  else if (mode==QUERY) {
    noBP = false;
    gap = -27;
    ext = -0.1;
    alpha = 1.0;
    beta = 1.0;
  }
  else {
    assert(true);
  }
}

void BPLAKernel::Options::
add_options(po::options_description& desc)
{
  desc.add_options()
    //("noBP", 
    // po::value<bool>(&noBP)->zero_tokens()->default_value(false),
    // "run as local alignment kernels")
    //("gap,g", 
    // po::value<double>(&gap)->default_value(-27), 
    // "set gap")
    //("ext,e", 
    // po::value<double>(&ext)->default_value(-0.1), 
    // "set ext")
    //("alpha,a", 
    // po::value<double>(&alpha)->default_value(1.0), 
    // "set alpha")
    ("beta,b", 
     po::value<double>(&beta)->default_value(0.1), 
     "set beta")
    ("score", 
     po::value<string>(&score_file), 
     "specify the score table file")
    ;
}

struct LAScore
{
  LAScore(const multi_array<double,2>& score_table) 
    : score_table_(score_table) {}

  inline double 
  operator()(const FeatureVector& x, const FeatureVector& y, 
	     uint i, uint j) const
  {
    double v = 0.0;
    float n = 0.0;
    for (uint k=0; k!=N_RNA; ++k) {
      if (x[i+k]==0.0) continue;
      for (uint l=0; l!=N_RNA; ++l) {
        if (y[j+l]==0.0) continue;
        n += x[i+k]*y[j+l];
        v += score_table_[k][l]*x[i+k]*y[j+l];
      }
    }
    return n==0.0 ? 0.0 : v/n;
  }

  const multi_array<double,2>& score_table_;
};

struct BPLAScore : public LAScore
{
  BPLAScore(const multi_array<double,2>& score_table, double alpha)
    : LAScore(score_table), alpha_(alpha) {}

  inline double
  operator()(const FeatureVector& x, const FeatureVector& y, 
	     uint i, uint j) const
  {
    return alpha_ * (x[i+FEAT_STR_L]*y[j+FEAT_STR_L] + x[i+FEAT_STR_R]*y[j+FEAT_STR_R])
      + x[i+FEAT_STR_U]*y[j+FEAT_STR_U] * LAScore::operator()(x,y,i,j);
  }

  double alpha_;
};

template < class Score > 
inline double 
local_alignment(const FeatureVector& x, const FeatureVector& y,
		double beta, double gap, double ext, Score score,
		bool rescale, double& scale)
{
  typedef multi_array<double,2> dp_type;
  const uint xsize = x.size()/N_FEAT;
  const uint ysize = y.size()/N_FEAT;

  scale = 1.0;
  double sf = 0.6;
  double ds = 1.0;

  if (rescale) {
    double viterbi = 0.0;
    dp_type M(extents[xsize+1][ysize+1]);
    dp_type X(extents[xsize+1][ysize+1]);
    dp_type Y(extents[xsize+1][ysize+1]);
    
    //initialize  
    for (uint i=0; i!=xsize+1; ++i) {
      M[i][0]=0;
      X[i][0]=0;
      Y[i][0]=0;
    }
    for (uint j=0; j!=ysize+1; ++j) {
      M[0][j]=0;
      X[0][j]=0;
      Y[0][j]=0;
    }
    
    //calculate
    for (uint i=1, xi=0; i != xsize+1; ++i, xi+=N_FEAT) {
      for (uint j=1, yj=0; j != ysize+1; ++j, yj+=N_FEAT) {
	vector<double> cmax(3, 0.0);
	vector<double> cval(3, 0.0);
	
	double sc = score(x,y,xi,yj);
	cval[0] = M[i-1][j-1] + sc;
	cval[1] = X[i-1][j-1] + sc;
	cval[2] = Y[i-1][j-1] + sc;
	cmax[0] = sc;
	for (uint k=0; k != cval.size(); ++k) 
	  if (cmax[0] < cval[k])
	    cmax[0] = cval[k];
	M[i][j] = cmax[0];
	
	cval[0] = M[i-1][j] + gap;
	cval[1] = X[i-1][j] + ext;
	cval[2] = Y[i-1][j] + ext;
	cmax[1] = cval[0];
	for (uint k=1; k != cval.size(); ++k) 
	  if (cmax[1] < cval[k])
	    cmax[1] = cval[k];
	X[i][j] = cmax[1];
	
	cval[0] = M[i][j-1] + gap;
	cval[1] = X[i][j-1] + ext;
	cval[2] = Y[i][j-1] + ext;
	cmax[2] = cval[0];
	for (uint k=1; k != cval.size(); ++k) 
	  if (cmax[2] < cval[k])
	    cmax[2] = cval[k];
	Y[i][j] = cmax[2];
	
	for (uint k=0; k != cmax.size(); ++k) 
	  if (viterbi < cmax[k])
	    viterbi = cmax[k];
      }
    }
    scale = exp(- sf*beta*viterbi);
    ds = exp(- sf*beta*viterbi / (xsize+ysize) );
  }

  dp_type M(extents[xsize+1][ysize+1]);
  dp_type X(extents[xsize+1][ysize+1]);
  dp_type Y(extents[xsize+1][ysize+1]);
  dp_type X2(extents[xsize+1][ysize+1]);
  dp_type Y2(extents[xsize+1][ysize+1]);
  vector<double> sone(xsize+ysize+1, 1.0);

  double beta_gap = exp(beta*gap);
  double beta_ext = exp(beta*ext);

  //initialize  
  for (uint i=0; i!=xsize+1; ++i) {
    M[i][0]=0;
    X[i][0]=0;
    Y[i][0]=0;
    X2[i][0]=0;
    Y2[i][0]=0;
  }
  for (uint j=0; j!=ysize+1; ++j) {
    M[0][j]=0;
    X[0][j]=0;
    Y[0][j]=0;
    X2[0][j]=0;
    Y2[0][j]=0;
  }
  for (uint i=1; i!=xsize+ysize+1; ++i) 
    sone[i] = sone[i-1]*ds;

  //calculate
  for (uint i=1, xi=0; i != xsize+1; ++i, xi+=N_FEAT) {
    for (uint j=1, yj=0; j != ysize+1; ++j, yj+=N_FEAT) {
      M[i][j] = exp(beta * score(x,y,xi,yj))
	* (sone[i+j-2] + X[i-1][j-1] + Y[i-1][j-1] + M[i-1][j-1]);

      X[i][j] = beta_gap * M[i-1][j] 
	+ beta_ext * X[i-1][j] ;
      Y[i][j] = beta_gap * (M[i][j-1] + X[i][j-1]) 
	+ beta_ext * Y[i][j-1] ;
      
      X2[i][j]= M[i-1][j] + X2[i-1][j];
      Y2[i][j]= M[i][j-1] + X2[i][j-1] + Y2[i][j-1];

      M[i][j] *= ds*ds;
      X[i][j] *= ds;
      Y[i][j] *= ds;
      X2[i][j] *= ds;
      Y2[i][j] *= ds;
    }
  }

  return sone[xsize+ysize] 
    + X2[xsize][ysize] + Y2[xsize][ysize] + M[xsize][ysize];
}

BPLAKernel::
BPLAKernel(const BPLAKernel::Options& opts) 
  : score_table_(extents[N_RNA][N_RNA]), noBP_(opts.noBP), 
    gap_(opts.gap), ext_(opts.ext), 
    alpha_(opts.alpha), beta_(opts.beta) 
{
  read_score_table(opts.score_file);
}

double BPLAKernel::
operator()(const FeatureVector& x, const FeatureVector& y, 
	   bool rescale, double& scale) const
{
  assert(x.size()%N_FEAT==0 && y.size()%N_FEAT==0);

  if (noBP_)
    return local_alignment(x, y, beta_, gap_, ext_, LAScore(score_table_),
			   rescale, scale);
  else
    return local_alignment(x, y, beta_, gap_, ext_, BPLAScore(score_table_, alpha_),
			   rescale, scale);
}

double BPLAKernel::
operator()(const FeatureVector& x, const FeatureVector& y) const
{
  double tmp;
  return (*this)(x,y,false,tmp);
}

void BPLAKernel::
read_score_table(const string& score_file)
{
  if (score_file.empty()) {
    for (uint i=0; i!=N_RNA; ++i) 
      for (uint j=0; j!=N_RNA; ++j) 
        score_table_[i][j] = default_score_table[i][j];
  }
  else {
    ifstream in(score_file.c_str());
    if (in.fail()) perror(score_file.c_str()); 
    string a, b;
    float v;
    while (in >> a >> b >> v) {
      score_table_[char2rna(a[0])][char2rna(b[0])] = v;
    }
  }
}
