#include <cmath>
#include <sstream>

#include "Data.h"
#include "kernel/common/rna.h"
#include "kernel/common/profile.h"
#include "kernel/common/fa.h"
#include "kernel/common/maf.h"
#include "kernel/common/aln.h"

using namespace std;
using namespace boost;

static void
fill_seq_weight(vector<FeatureVector>& feat, 
		const FeatureSequence::IS& seq)
{
  ProfileSequence s(seq);
  feat.resize(s.size(),FeatureVector(N_FEAT,0.0));

  for (int i=0; i!=feat.size(); ++i) {
    double n = 0.0;
    for (int j=0; j!=N_RNA; ++j) 
      n += s[i][j];
    for (int j=0; j!=N_RNA; ++j) 
      feat[i][j] = s[i][j] / n;
  }
}

static void
fill_str_weight(vector<FeatureVector>& feat, 
		const FeatureSequence::IS& seq, 
		float pf_scale, const BPMatrix::Options& opts)
{
  BPMatrix bp(seq, pf_scale, opts);
  assert(bp.size()==feat.size());
  for (int i=0; i<bp.size(); ++i) {
    feat[i][FEAT_STR_L] = feat[i][FEAT_STR_R] = 0.0;
    for (int j=0; j<i; ++j)
      feat[i][FEAT_STR_R] += bp(j+1, i+1);
    for (int j=i+1; j<bp.size(); ++j) 
      feat[i][FEAT_STR_L] += bp(i+1, j+1);
    feat[i][FEAT_STR_U] = 1.0-(feat[i][FEAT_STR_L]+feat[i][FEAT_STR_R]);
    if (feat[i][FEAT_STR_U] < 0.0) feat[i][FEAT_STR_U] = 0.0; // error around
    feat[i][FEAT_STR_R] = sqrt(feat[i][FEAT_STR_R]);
    feat[i][FEAT_STR_L] = sqrt(feat[i][FEAT_STR_L]);
    feat[i][FEAT_STR_U] = sqrt(feat[i][FEAT_STR_U]);
  }
}

FeatureSequence::
FeatureSequence(const FeatureSequence::IS& seq)
{
  fill_seq_weight(feat,seq);
}

FeatureSequence::
FeatureSequence(const FeatureSequence::IS& seq, 
		float pf_scale, const BPMatrix::Options& opts)
{
  fill_seq_weight(feat,seq);
  fill_str_weight(feat,seq,pf_scale,opts);
}
 
enum { TP_UNKNOWN, TP_FA, TP_ALN, TP_MAF };

static int
check_filetype(const char* f)
{
  ifstream in(f);
  string l;
  while (getline(in, l)) {
    if (l[0]=='>') return TP_FA;
    else if (l.compare(0, 7, "CLUSTAL")==0) return TP_ALN;
    else if (l.compare(0, 2, "a ")==0) return TP_MAF;
  }
  return TP_UNKNOWN;
}

DataLoader::
DataLoader(const char* filename, 
	   const BPMatrix::Options& bp_opts, bool use_bp)
  : bp_opts_(bp_opts), use_bp_(use_bp), 
    filename_(filename), type_(check_filetype(filename)), 
    fi_(), pf_is_(NULL)
{
  switch (type_) {
  case TP_FA:
  case TP_ALN:
  case TP_MAF:
    fi_ = boost::spirit::file_iterator<>(filename_);
    break;
  default:
    break;
  }
  if (!fi_) {
    ostringstream os;
    os << filename_ << ": no such file";
    throw os.str().c_str();
    //return false;
  }
}

DataLoader::
DataLoader(const char* filename, const char* pf_scales,
	   const BPMatrix::Options& bp_opts, bool use_bp)
  : bp_opts_(bp_opts), use_bp_(use_bp),
    filename_(filename), type_(check_filetype(filename)),
    fi_(), pf_is_(NULL)
{
  switch (type_) {
  case TP_FA:
  case TP_ALN:
  case TP_MAF:
    fi_ = boost::spirit::file_iterator<>(filename_);
    break;
  default:
    break;
  }
  if (!fi_) {
    ostringstream os;
    os << filename_ << ": no such file";
    throw os.str().c_str();
    //return false;
  }

  pf_is_ = new ifstream(pf_scales);
}

DataLoader::
~DataLoader()
{
  if (pf_is_) delete pf_is_;
}

DataLoader::Data* DataLoader::
get()
{
  bool ret=false;
  list<string> ma;
  switch (type_) {
  case TP_FA:
    ret=load_fa(ma, fi_);
    break;
  case TP_ALN:
    ret=load_aln(ma, fi_);
    break;
  case TP_MAF:
    ret=load_maf(ma, fi_);
    break;
  default:
    break;
  }

  float pf_scale = -1;
  if (pf_is_ && *pf_is_) {
    *pf_is_ >> pf_scale;
    //std::cout << pf_scale << std::endl;
  }

  if (ret) {
    list<string>::const_iterator x;
    int l=ma.begin()->size();
    for (x=ma.begin(); x!=ma.end(); ++x) {
      if (l!=x->size()) throw "wrong alignment";
    }
  
    if (use_bp_)
      return new Data(ma, pf_scale, bp_opts_);
    else
      return new Data(ma);
  }
  return NULL;
}
