#include <string>
#include <queue>
#include <cassert>

#include "ForestLSH.h"
#include "KernelCalculator.h"

using namespace std;
using namespace boost;

#define THSH_SIZE 20
#define LOSS_SIZE 10
#define MAX_NPROB 20
const double thsh_[THSH_SIZE] = {0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40, 0.45, 
				 0.50, 0.55, 0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95};
const double loss_[LOSS_SIZE] = {0.01, 0.03, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40, 0.50, 0.60}; 

//---------- LSH Forest with Self-Tuning of Parameters ------------------
//---------- Bawa et al, 2005

template <class K>
ForestLSH<K>::
ForestLSH(const LSH::Options& opts, 
	  uint ntrn, uint nsmp, bool rescale, const K& kernel)
  : LSH(opts,true), 
    ntrn_(ntrn), nsmp_(nsmp), rescale_(rescale), kernel_(kernel)
{
  // do not use constructor tree_(ntab,new HashNode()) 
  // since it causes all tree_[i] have the same address
  tree_.resize(ntab_);
  for (uint i=0; i!=ntab_; ++i) 
    tree_[i] = new HashNode();

  assert(nsec_ == 2);
  assert(nsmp_ <= ntrn_);
}

template <class K>
void ForestLSH<K>::
make_function(uint nthr)
{
  assert(ntrn_ <= data_.size());

  trn_.resize(ntrn_);

  deque<bool> pick(data_.size(),false); // do not use vector<bool> broken in STL
  RandomShuffler rns(rng_);

  fill_n(pick.begin(),ntrn_,true);
  random_shuffle(pick.begin(),pick.end(),rns);

  for (uint p=0, i=0; i!=data_.size(); ++i) 
    if (pick[i]) trn_[p++] = get_feat(i);

  /* selection with replacements
  uniform_int<> unif_dist(0,data_.size()-1);
  variate_generator<RandomGenerator&,uniform_int<> > unif_samp(rng_,unif_dist);

  for (uint i=0; i!=ntrn_; ++i) 
    trn_[i] = get_feat(unif_samp());
  */

  KernelSearch sh;
  sh.prepare(trn_,kernel_,rescale_,nthr);
  sh.get_db(trn_score_,trn_scale_);

  KernelMatrix mx;
  mx.calculate(trn_,kernel_,false,normalize_,rescale_,nthr);
  trn_matrix_.resize(ntrn_,vector<double>(ntrn_));
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      trn_matrix_[i][j] = mx(i,j);

  hf_ = KernelizedLSH<K>::hash_function(trn_matrix_,ntrn_,nsmp_,ntab_,nand_,rng_);
}

template <class K>
void ForestLSH<K>::
make_index(uint nthr)
{
  for (uint i=0; i!=data_.size(); ++i) {
    vector<double> ks(ntrn_);
    KernelSearch sh;
    sh.set_db(trn_score_,trn_scale_);
    sh.calculate(get_feat(i),trn_,kernel_,false,normalize_,rescale_,nthr);
    for (uint j=0; j!=ntrn_; ++j)
      ks[j] = sh(j);

    for (uint j=0; j!=ntab_; ++j) {
      HashPath s;
      uint dv = 0;
      for (uint k=0; k!=nand_; ++k) {
	double g = inner_product(ks.begin(),ks.end(),hf_[j][k].begin(),0.0);
	if (g > 0) s.push_back(true);
	else s.push_back(false);
      }
      HashNode* in = descend(s,dv,tree_[j]);
      insert(i,s,dv,in);
    }
  }

  trunc();
}

template <class K>
void ForestLSH<K>::
print_function() 
{
  for (uint i=0; i!=ntab_; ++i) {
    cout << "LSH function " << i << ":" << endl << flush;
    for (uint j=0; j!=nand_; ++j) {
      cout << "hyper plane " << j << ":" << endl << flush;
      for (uint k=0; k!=ntrn_; ++k) {
	cout << hf_[i][j][k] << " ";
      }
      cout << endl << flush;
    }
  }
}

template <class K>
void ForestLSH<K>::
print_index() 
{
  cout << "ntab:" << ntab_ << " nand:" << nand_ << " nsec:" << nsec_ 
       << " normalize:" << normalize_ << " rescale:" << rescale_ 
       << endl << flush;
  cout << "ntrn:" << ntrn_ << " nsmp:" << nsmp_
       << endl << flush;

  cout << "thsh: " << endl << flush;
  for (uint i=0; i!=THSH_SIZE; ++i) 
    cout << thsh_[i] << " ";
  cout << endl << flush;

  cout << "loss: " << endl << flush;
  for (uint i=0; i!=LOSS_SIZE; ++i) 
    cout << loss_[i] << " ";
  cout << endl << flush;

  cout << "opt_nprb: " << endl << flush;
  for (uint i=0; i!=THSH_SIZE; ++i) {
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      cout << opt_nprb_[i][j] << " ";
    cout << endl << flush;
  }
  
  for (uint i=0; i!=ntab_; ++i) {
    cout << "LSH tree " << i << ":" << endl << flush;
    HashPath head;
    print_node(tree_[i], head);
  }
}

template <class K>
void ForestLSH<K>::
clear()
{
  vector<FeatureVector>().swap(mother_);
  vector<HashData>().swap(data_);
  for (uint i=0; i!=ntab_; ++i) 
    clear_node(tree_[i]);
}

template <class K>
void ForestLSH<K>::
probe(vector<HashHit>& hit, const FeatureVector& x,
      uint nprb, uint nthr)
{
  vector<double> ks(ntrn_);
  KernelSearch sh;
  sh.set_db(trn_score_,trn_scale_);
  sh.calculate(x,trn_,kernel_,false,normalize_,rescale_,nthr);
  for (uint i=0; i!=ntrn_; ++i)
    ks[i] = sh(i);

  probe(hit,ks,nprb);
}

template <class K>
void ForestLSH<K>::
probe(vector<HashHit>& hit, const vector<double>& ks,
      uint nprb)
{
  assert(ks.size()==ntrn_);

  list<HashHit> tmp; // use list for massive push_back()

  for (uint i=0; i!=ntab_; ++i) {
    vector<HashNode*> seq;
    multi_probe(seq,ks,i,nprb);
    for (uint j=0; j!=seq.size(); ++j) {
      HashBacket::iterator k = seq[j]->b.begin();
      HashBacket::iterator ke = seq[j]->b.end();
      for ( ; k!=ke; ++k) {
	HashHit h;
	h.i = *k;
	h.s = 0.0;
	tmp.push_back(h); 
      }
    }
  }

  hit.assign(tmp.begin(),tmp.end());
  sort(hit.begin(),hit.end(),HashHitLess());
  vector<HashHit>::iterator he = unique(hit.begin(),hit.end(),HashHitSame());
  hit.erase(he,hit.end());
}

template <class K>
void ForestLSH<K>::
optimize_nprb(uint nthr)
{
  // resample virtual queries
  vector<FeatureVector> vq(ntrn_);
  vector<vector<double> > vq_ks(ntrn_,vector<double>(ntrn_));
  KernelSearch vq_sh;

  deque<bool> pick(data_.size(),false); // do not use vector<bool> broken in STL
  RandomShuffler rns(rng_);

  fill_n(pick.begin(),ntrn_,true);
  random_shuffle(pick.begin(),pick.end(),rns);

  for (uint p=0, i=0; i!=data_.size(); ++i) 
    if (pick[i]) vq[p++] = get_feat(i);

  vq_sh.prepare(trn_,kernel_,rescale_,nthr);
  vq_sh.calculate(vq,trn_,kernel_,false,normalize_,rescale_,nthr);
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      vq_ks[i][j] = vq_sh(i,j);

  // edge[i][j] = k means that 
  // trn_[i] captures data_[j] only by the threshold < thsh_[k] 
  vector<vector<double> > kv(ntrn_,vector<double>(data_.size()));
  vector<vector<uint> > edge(ntrn_,vector<uint>(data_.size(),THSH_SIZE));

  KernelSearch qu_sh;
  vector<double> score;
  vector<double> scale;
  typename K::Options qu_opts(K::QUERY);
  K qu_kernel(qu_opts);
  bool qu_logarithm = true;
  bool qu_rescale = true;
  qu_sh.prepare(vq,qu_kernel,qu_rescale,nthr);
  qu_sh.get_db(score,scale);

  for (uint i=0; i!=data_.size(); ++i) {
    KernelSearch sh;
    sh.set_db(score,scale); 
    sh.calculate(get_feat(i),vq,qu_kernel,qu_logarithm,true,qu_rescale,nthr);
    for (uint j=0; j!=ntrn_; ++j)
      kv[j][i] = sh(j);
  }
  
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=data_.size(); ++j) 
      for (uint k=0; k!=THSH_SIZE; ++k) 
	if (kv[i][j] < thsh_[k]) {
	  edge[i][j] = k;
	  break;
	}

  // nhit[i][j] = f means that
  // for i+1 multi-probe and thsh_[j], fraction f hits are missed
  vector<uint> nexact(THSH_SIZE,0);
  vector<vector<double> > fmiss(MAX_NPROB,vector<double>(THSH_SIZE,0.0));

  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      for (uint k=0; k!=data_.size(); ++k) 
	if (i < edge[j][k]) nexact[i]++;

  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=ntrn_; ++j) {
      vector<HashHit> hit;
      probe(hit,vq_ks[j],i+1);
      for (uint k=0; k!=hit.size(); ++k)
	for (uint l=0; l!=edge[j][hit[k].i]; ++l)
	  fmiss[i][l] += 1.0;
    }
  }
  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=THSH_SIZE; ++j) {
      if (nexact[j]==0) fmiss[i][j] = 0.0;
      else fmiss[i][j] = 1.0 - fmiss[i][j] / nexact[j];
    }
  }

  // opt_probe_[i][j] = k means that 
  // for thsh_[i] and loss_[j], k is the optimized nprb 
  opt_nprb_.assign(THSH_SIZE,vector<uint>(LOSS_SIZE,MAX_NPROB));
  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      for (uint k=0; k!=MAX_NPROB; ++k) 
	if (fmiss[k][i] < loss_[j]) {
	  opt_nprb_[i][j] = k+1;
	  break;
	}

  // use trn_ as virtual queries
  /* 
  // edge[i][j] = k means that 
  // trn_[i] captures data_[j] only by the threshold < thsh_[k] 
  vector<vector<double> > kv(ntrn_,vector<double>(data_.size()));
  vector<vector<uint> > edge(ntrn_,vector<uint>(data_.size(),THSH_SIZE));

  KernelSearch qu_sh;
  vector<double> score;
  vector<double> scale;
  typename K::Options qu_opts(K::QUERY);
  K qu_kernel(qu_opts);
  bool qu_logarithm = true;
  bool qu_rescale = true;
  qu_sh.prepare(trn_,qu_kernel,qu_rescale,nthr);
  qu_sh.get_db(score,scale);

  for (uint i=0; i!=data_.size(); ++i) {
    KernelSearch sh;
    sh.set_db(score,scale); 
    sh.calculate(get_feat(i),trn_,qu_kernel,qu_logarithm,true,qu_rescale,nthr);
    for (uint j=0; j!=ntrn_; ++j)
      kv[j][i] = sh(j);
  }
  
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=data_.size(); ++j) 
      for (uint k=0; k!=THSH_SIZE; ++k) 
	if (kv[i][j] < thsh_[k]) {
	  edge[i][j] = k;
	  break;
	}

  // nhit[i][j] = f means that
  // for i+1 multi-probe and thsh_[j], fraction f hits are missed
  vector<uint> nexact(THSH_SIZE,0);
  vector<vector<double> > fmiss(MAX_NPROB,vector<double>(THSH_SIZE,0.0));

  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      for (uint k=0; k!=data_.size(); ++k) 
	if (i < edge[j][k]) nexact[i]++;

  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=ntrn_; ++j) {
      vector<HashHit> hit;
      probe(hit,trn_matrix_[j],i+1);
      for (uint k=0; k!=hit.size(); ++k)
	for (uint l=0; l!=edge[j][hit[k].i]; ++l)
	  fmiss[i][l] += 1.0;
    }
  }
  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=THSH_SIZE; ++j) {
      if (nexact[j]==0) fmiss[i][j] = 0.0;
      else fmiss[i][j] = 1.0 - fmiss[i][j] / nexact[j];
    }
  }

  // opt_probe_[i][j] = k means that 
  // for thsh_[i] and loss_[j], k is the optimized nprb 
  opt_nprb_.assign(THSH_SIZE,vector<uint>(LOSS_SIZE,MAX_NPROB));
  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      for (uint k=0; k!=MAX_NPROB; ++k) 
	if (fmiss[k][i] < loss_[j]) {
	  opt_nprb_[i][j] = k+1;
	  break;
	}
  */
}

template <class K>
uint ForestLSH<K>::
estimate_nprb(double thsh, double loss) {
  uint i, j;

  for (i=0; i!=THSH_SIZE; ++i) 
    if (thsh < thsh_[i]) break;
  if (i==0) return MAX_NPROB;

  for (j=0; j!=LOSS_SIZE; ++j) 
    if (loss < loss_[j]) break;
  if (j==0) return MAX_NPROB;

  return opt_nprb_[i-1][j-1];
}

template <class K>
typename ForestLSH<K>::HashNode* ForestLSH<K>::
descend(HashPath& s, uint& dv, HashNode* rt)
{
  HashNode* n = rt;

  while (!s.empty()) {
    for ( ; dv!=n->s.size(); ++dv) {
      if (n->s[dv]==s.front()) s.pop_front();
      else return n; 
    }
    if (s.empty()) break;

    dv = 0;
    if (s.front()) {
      if (n->rc==NULL) {
	n->cnct = true;
	return n;
      }
      else {
	n = n->rc;
      }
    }
    else {
      if (n->lc==NULL) {
	n->cnct = true;
	return n;
      }
      else {
	n = n->lc;
      }
    }
    s.pop_front();
  }

  n->term = true;
  return n;
}

template <class K>
void ForestLSH<K>::
insert(uint idx, HashPath& s, uint& dv, HashNode* in)
{
  if (in->connectable()) { 
    assert(!s.empty());
    assert( (s.front() && in->rc==NULL) || 
            (!s.front() && in->lc==NULL) );

    HashNode* n = new HashNode();
    if (s.front()) in->rc = n;
    else in->lc = n;
    in->cnct = false;
    s.pop_front();
    n->b.push_back(idx);
    n->s = s;
    n->p = in;

    return;
  }

  if (!in->terminal()) {
    assert(!s.empty());
    assert(s.size()+dv >= in->s.size());

    HashNode* h = new HashNode();
    if (in->p->rc==in) in->p->rc = h;
    else in->p->lc = h;
    h->p = in->p;

    for (uint i=0; i!=dv; ++i) {
      h->s.push_back(in->s.front());
      in->s.pop_front();
    }
    
    HashNode* n = new HashNode();
    if (in->s.front()) {
      h->rc = in;
      h->lc = n;
    }
    else {
      h->rc = n;
      h->lc = in;
    }
    in->s.pop_front();
    in->p = h;
    s.pop_front();
    n->b.push_back(idx);
    n->s = s;
    n->p = h;

    return;
  }

  assert(s.empty());

  in->term = false;
  in->b.push_back(idx);
}

template <class K>
void ForestLSH<K>::
print_node(HashNode* n, HashPath& h) 
{
  for (uint i=0; i!=h.size(); ++i) 
    cout << h[i];
  if (h.empty()) {
    assert(n->p==NULL);
    cout << ">ROOT" << endl << flush;
  }
  else if (h.back()) {
    assert(n->p->rc==n);
    cout << ">R-CHILD" << endl << flush;
  }
  else if (!h.back()) {
    assert(n->p->lc==n);
    cout << ">L-CHILD" << endl << flush;
  }

  HashBacket::iterator k = n->b.begin();
  HashBacket::iterator ke = n->b.end();
  for ( ; k!=ke; ++k) {
    const HashData& d = get_data(*k);
    
    for (uint i=0; i!=h.size(); ++i) 
      cout << h[i];
    cout << ">(" 
	 << d.sid_ << "," << d.pos_ << "," 
	 << d.win_ << "," << d.off_ 
	 << ")" << endl << flush; 
  }

  uint ex = n->s.size();

  for (uint i=0; i!=ex; ++i) 
    h.push_back(n->s[i]); 
  for (uint i=0; i!=h.size(); ++i) 
    cout << h[i];
  cout << ">EXPAND" << endl << flush;

  if (n->rc!=NULL) {
    assert(n->b.empty());
    for (uint i=0; i!=h.size(); ++i) 
      cout << h[i];  
    cout << ">TO-R" << endl << flush;
    h.push_back(true);
    print_node(n->rc,h);
    h.pop_back();
  }
  if (n->lc!=NULL) {
    assert(n->b.empty());
    for (uint i=0; i!=h.size(); ++i) 
      cout << h[i];  
    cout << ">TO-L" << endl << flush;
    h.push_back(false);
    print_node(n->lc,h);
    h.pop_back();
  }
  if (n->lc==NULL && n->rc==NULL) {
    assert(!n->b.empty());
    for (uint i=0; i!=h.size(); ++i) 
      cout << h[i];
    cout << ">STOP" << endl << flush;
  }

  for (uint i=0; i!=ex; ++i) 
    h.pop_back(); 
}

// free nodes in a post order traversal
template <class K>
void ForestLSH<K>::
clear_node(HashNode* n)
{
  if (n!=NULL) {
    clear_node(n->lc);
    clear_node(n->rc);
    delete n;
  }
  return;
}

template <class K>
void ForestLSH<K>::
trunc()
{
  for (uint i=0; i!=ntab_; ++i) 
    trunc_node(tree_[i]);
}

template <class K>
void ForestLSH<K>::
trunc_node(HashNode* n)
{
  if (n!=NULL) {
    if (n->leaf()) n->s.clear();
    trunc_node(n->lc);
    trunc_node(n->rc);
  }
  return;
}

template <class K>
void ForestLSH<K>::
multi_probe(vector<HashNode*>& seq, 
	    const vector<double>& ks, uint idx, uint nprb)
{
  list<HashNode*> tmp; // use list for massive push_back()
  HashPath sfx;
  uint dv = 0;

  for (uint i=0; i!=nand_; ++i) {
    double g = inner_product(ks.begin(),ks.end(),hf_[idx][i].begin(),0.0);
    if (g > 0) sfx.push_back(true);
    else sfx.push_back(false);
  }
  HashNode* n = descend(sfx,dv,tree_[idx]);
  HashNode* c = n;

  if (n->leaf()) {
    tmp.push_back(n);
    n = n->p;
  }

  for (uint i=1; i!=nprb; ++i) {
    if (n==NULL) break;

    if (n->rc==c) bf_probe(tmp,n->lc);
    else bf_probe(tmp,n->rc);
    c = n;
    n = n->p;
  }

  seq.assign(tmp.begin(),tmp.end());
}

// probe a subtree by breadth first search
template <class K>
bool ForestLSH<K>::
bf_probe(list<HashNode*>& prb, HashNode* rt)
{
  queue<HashNode*> q;
  q.push(rt);

  while (!q.empty()) {
    HashNode* n = q.front();
    q.pop();

    if (n==NULL) continue;

    if (n->leaf()) prb.push_back(n);
    q.push(n->rc);
    q.push(n->lc);
  }

  return false;
}

// instantiation

#include "BPLAKernel.h"
template class ForestLSH<BPLAKernel>;
