#ifndef __INC_FOREST_LSH_H__
#define __INC_FOREST_LSH_H__

#include <vector>
#include <list>
#include <deque>

#include <limits.h>  // necessary for boost/serialization
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/deque.hpp>

#include "Utility.h"
#include "LSH.h"
#include "KernelizedLSH.h"

//---------- LSH Forest with Self-Tuning of Parameters ------------------
//---------- Bawa et al, 2005

template <class K>
class ForestLSH : public LSH
{
public:
  typedef std::deque<bool> HashPath; 

private:
  struct HashNode {
    HashNode() : p(NULL), rc(NULL), lc(NULL), cnct(false), term(false) {}
    HashBacket b;
    HashPath s;
    HashNode* p; 
    HashNode* rc;
    HashNode* lc;
    bool cnct;
    bool term;

    bool connectable() 
    { 
      return cnct;
    }
    bool leaf() 
    { 
      return !b.empty() && p!=NULL && rc==NULL && lc==NULL;
    }
    bool terminal() 
    { 
      return term;
    }

  private:
    friend class boost::serialization::access;
    template <class Archive> 
    void serialize(Archive& ar, const uint ver) 
    {
      ar & b & s;
      ar & p & lc & rc; 
    }
  };

  typedef HashNode* HashTree;

public:
  ForestLSH(const LSH::Options& opts, 
	    uint ntrn, uint nsmp, bool rescale, const K& kernel);
  ForestLSH() : LSH(), ntrn_(0), nsmp_(0), rescale_(false), kernel_() {}
  ~ForestLSH() {}

public:
  void make_function(uint nthr);
  void make_index(uint nthr);
  void print_function();
  void print_index();
  void clear();
  void probe(std::vector<HashHit>& hit, const FeatureVector& x, 
	     uint nprb, uint nthr);
  void probe(std::vector<HashHit>& hit, const std::vector<double>& ks, 
	     uint nprb);
  void optimize_nprb(uint nthr);
  uint estimate_nprb(double thsh, double loss);

private:
  HashNode* descend(HashPath& s, uint& dv, HashNode* rt);
  void insert(uint idx, HashPath& s, uint& dv, HashNode* in);

  void print_node(HashNode* n, HashPath& h);
  void clear_node(HashNode* n);

  void trunc();
  void trunc_node(HashNode* n);

  void multi_probe(std::vector<HashNode*>& seq, 
		   const std::vector<double>& ks, uint idx, uint nprb);

  bool bf_probe(std::list<HashNode*>& prb, HashNode* rt);

  // pseudo override for pure virtual functions in class LSH
  // these functions are unnecessary or overloaded in this child class
  HashValue hash_value(const FeatureVector& x, uint idx) { return 0; }
  void multi_probe(std::vector<HashValue>& seq, 
		   const FeatureVector& x, uint idx, uint nprb) {} 

private:
  std::vector<HashTree> tree_;
  std::vector<typename KernelizedLSH<K>::HashFunction> hf_;
  std::vector<FeatureVector> trn_;
  std::vector<double> trn_score_;
  std::vector<double> trn_scale_;
  std::vector<std::vector<double> > trn_matrix_;
  std::vector<std::vector<uint> > opt_nprb_; 
  const uint ntrn_;
  const uint nsmp_;
  const bool rescale_;
  const K kernel_;

private:
  friend class boost::serialization::access;
  template <class Archive> 
  void serialize(Archive& ar, const uint ver)
  {
    ar & boost::serialization::base_object<LSH>(*this);
    ar & tree_; 
    ar & hf_ & trn_ & trn_score_ & trn_scale_ & opt_nprb_; 
    ar & const_cast<uint&>(ntrn_) & const_cast<uint&>(nsmp_);
    ar & const_cast<bool&>(rescale_) & const_cast<K&>(kernel_);
  }
};

#endif
