#ifndef __INC_FRAMEWORK_KIND_H__
#define __INC_FRAMEWORK_KIND_H__

#include <vector>
#include <list>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <boost/program_options.hpp>
#include <boost/timer.hpp>

#include <limits.h>  // necessary for boost::serialization
#include <boost/serialization/serialization.hpp>
#include <boost/archive/binary_oarchive.hpp>

#include "Utility.h"
#include "LSH.h"
#include "Data.h"
#include "kernel/common/glob_wrapper.h"

struct Options
{
  // inputs
  std::vector<std::string> seq_files;
  // outputs
  std::string output;
  // opts
  uint nthr;
  uint ntrn;
  uint nsmp;
  static const bool rescale = true;
  static const bool forest = true;
  bool verbose;

  Options() : seq_files(), output() {}

  void add_options(boost::program_options::options_description& opts);
  void parse_extra_args(const std::vector<std::string>& extra_args);
};

template < class LSH, class LDF >
class App
{
public:
  typedef typename LDF::Data Data;

public:
  App(LSH& lsh, const LDF& ldf, const Options& opts)
    : lsh_(lsh), ldf_(ldf), opts_(opts) {}

  bool execute() 
  {
    return create();
  }

private:
  bool create() 
  {
    std::cout << "Kind - create a kernel index database" << std::endl; 
    //std::cout << std::flush;

    std::cout << "load sequences" << std::endl;
    std::vector<uint> len;
    bool res = load_lsh_mother(len, opts_.seq_files);
    if (!res) return false;
    std::cout << "loaded " << len.size() << " sequences" << std::endl;
    //std::cout << std::flush;

    std::cout << "store fragments" << std::endl;
    uint cnt = load_lsh_data(len);
    if (!cnt) return false;
    std::cout << "stored " << cnt << " fragments" << std::endl;
    //std::cout << std::flush;

    std::cout << "make functions" << std::endl;
    lsh_.make_function(opts_.nthr);
    if (opts_.verbose) lsh_.print_function();
    //std::cout << std::flush;

    std::cout << "make indexes" << std::endl;
    lsh_.make_index(opts_.nthr);
    lsh_.optimize_nprb(opts_.nthr);
    if (opts_.verbose) lsh_.print_index();
    //std::cout << std::flush;

    std::cout << "output databases" << std::endl;
    std::string output_lsh = opts_.output + std::string(".lsh");
    std::ofstream ofs(output_lsh.c_str());
    boost::archive::binary_oarchive oar(ofs);
    oar << (const LSH&) lsh_;
    std::cout << "output " << output_lsh << std::endl;
    //std::cout << std::flush;

    std::cout << "Kind - finished" << std::endl;
    lsh_.clear();
    //std::cout << std::flush;

    return true;
  }

  bool load_lsh_mother(std::vector<uint>& len, 
		       const std::vector<std::string>& files) 
  {
    for (uint i=0; i!=files.size(); ++i) {
      Glob glob(files[i].c_str());
      if (glob.empty()) {
	std::ostringstream os;
	os << files[i] << ": no matches found";
	throw os.str().c_str();
	//return false;
      }
      Glob::const_iterator p;
      for (p=glob.begin(); p!=glob.end(); ++p) {
	double elapsed = 0.0;
	typename LDF::Loader* loader=ldf_.get_loader(p->c_str());
	if (loader==NULL) return false;
#ifdef HAVE_MPI
	if (MPI::COMM_WORLD.Get_rank()==0) {
#endif
	  std::cout << "loading " << *p << std::flush;
#ifdef HAVE_MPI
	}
#endif
	while (true) {
	  boost::timer tm;
	  Data* d = loader->get();
	  elapsed += tm.elapsed();
	  if (d==NULL) break;
	  lsh_.add_feat(d->feat);
	  len.push_back(d->size());
	  delete d;
	}
#ifdef HAVE_MPI
	if (MPI::COMM_WORLD.Get_rank()==0) {
#endif
	  std::cout << " (" << elapsed << "s) done." << std::endl;
#ifdef HAVE_MPI
	}
#endif
	delete loader;
      }
    }
    return true;
  }

  uint load_lsh_data(const std::vector<uint>& len)
  {
    uint sid=0, pos=0, off=0;
    uint cnt=0;

    while (true) {
      uint win = len[sid];
      uint slide = len[sid];

      while (pos+win > len[sid]) {
	off += len[sid] - pos;
	pos = 0;
	++sid;
	if (sid==len.size()) return cnt;
	win = len[sid];
	slide = len[sid];
      }

      if (opts_.verbose)
	std::cout << "storing " << "(" 
		  << sid << "," << pos << "," 
		  << win << "," << off << ")" 
		  << std::endl << std::flush; 
      
      HashData hd(sid, pos, win, off);
      lsh_.add_data(hd);
      ++cnt;

      off += slide;
      pos += slide;
    }

    return 0;
  }

private:
  LSH& lsh_;
  const LDF& ldf_;
  const Options& opts_;
};

#ifdef HAVE_MPI
class MPIState
{
public:
  MPIState(int& argc, char**& argv)
  {
    MPI::Init(argc, const_cast<char**&>(argv));
  }

  ~MPIState()
  {
    if (MPI::Is_initialized())
      MPI::Finalize();
  }
};
#endif

#endif	
