#ifndef __INC_FRAMEWORK_KIND_HOST_H__
#define __INC_FRAMEWORK_KIND_HOST_H__

#include <vector>
#include <list>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <boost/program_options.hpp>
#include <boost/timer.hpp>

#include <limits.h>  // necessary for boost::serialization
#include <boost/serialization/serialization.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include "Utility.h"
#include "LSH.h"
#include "Data.h"
#include "KernelCalculator.h"
#include "kernel/common/glob_wrapper.h"

struct Options
{
  // inputs
  std::string database;
  std::vector<std::string> seq_files;
  // outputs
  std::string output;
  // opts
  uint nthr;
  static const uint nprb = 0;
  double thsh;
  double loss;
  static const bool logarithm = true;
  static const bool rescale = true;
  static const bool forest = true;
  bool verbose;

  Options() : database(), seq_files(), output() {}

  void add_options(boost::program_options::options_description& opts);
  void parse_extra_args(const std::vector<std::string>& extra_args);
};

template < class LSH, class LDF, class KNL >
class App
{
public:
  typedef typename LDF::Data Data;

public:
  App(LSH& lsh, const LDF& ldf, const KNL& knl, const Options& opts)
    : lsh_(lsh), ldf_(ldf), knl_(knl), opts_(opts) {}

  bool execute() 
  {
    return search();
  }

private:
  bool search() 
  {
    std::cout << "KindHost - search a kernel index database" << std::endl; 
    //std::cout << std::flush;

    std::vector<FeatureVector> qu_mother;
    std::vector<HashData> qu_data;
    uint nprb = opts_.nprb;

    std::cout << "load databases" << std::endl;
    std::string database_lsh = opts_.database + std::string(".lsh");
    std::ifstream ifs(database_lsh.c_str());
    boost::archive::binary_iarchive iar(ifs);
    iar >> lsh_;
    if (nprb==0) nprb = lsh_.estimate_nprb(opts_.thsh,opts_.loss);
    if (opts_.verbose) lsh_.dump();
    //std::cout << std::flush;

    std::cout << "load sequences" << std::endl;
    std::vector<uint> len;
    bool res = load_lsh_mother(qu_mother, len, opts_.seq_files);
    if (!res) return false;
    std::cout << "loaded " << len.size() << " sequences" << std::endl;
    //std::cout << std::flush;

    std::cout << "store queries" << std::endl;
    uint cnt = load_lsh_data(qu_data, len);
    if (!cnt) return false;
    std::cout << "stored " << cnt << " queries" << std::endl;
    //std::cout << std::flush;

    std::cout << "search databases" << std::endl;
    std::string output_res = opts_.output + std::string(".res");
    std::ofstream ofs(output_res.c_str());

    for (uint i=0; i!=qu_data.size(); ++i) {
      std::vector<HashHit> hit; 
      KernelSearch sh;
      FeatureVector qu;
      std::vector<FeatureVector> db;

      std::cout << "query " << i;
      qu.resize(N_FEAT*qu_data[i].win_);
      const uint begin = qu_data[i].off_;
      const uint end = begin + qu_data[i].win_;
      for (uint p=0, j=begin; j!=end; ++j)
	for (uint k=0; k!=N_FEAT; ++k)
	  qu[p++] = qu_mother[j][k];
      std::cout << "; ";
      //std::cout << std::flush;

      std::cout << "probing " << nprb;
      lsh_.probe(hit,qu,nprb,opts_.nthr);
      db.resize(hit.size());
      for (uint j=0; j!=hit.size(); ++j) 
	db[j] = lsh_.get_feat(hit[j].i);
      std::cout << "; ";
      //std::cout << std::flush;

      std::cout << "reranking";
      sh.prepare(db,knl_,opts_.rescale,opts_.nthr);
      sh.calculate(qu,db,knl_,opts_.logarithm,true,opts_.rescale,opts_.nthr);
      for (uint j=0; j!=hit.size(); ++j) 
	hit[j].s = sh(j);
      std::sort(hit.begin(),hit.end(),HashHitScoreLarger());
      std::cout << "; ";
      //std::cout << std::flush;

      std::cout << "output";
      ofs << "KindHost search results " << i << std::endl;
      for (uint j=0; j!=hit.size(); ++j) {
	if (hit[j].s < opts_.thsh) break;
	const HashData& d = lsh_.get_data(hit[j].i);
	ofs << "query " << i << " "  << "("
	    << qu_data[i].sid_ << "," << qu_data[i].pos_ << "," 
	    << qu_data[i].win_ << "," << qu_data[i].off_ << ")" 
	    << " to fragment " << "(" 
	    << d.sid_ << "," << d.pos_ << "," 
	    << d.win_ << "," << d.off_ << ")"
	    << " " << hit[j].s  
	    << std::endl;
      }
      ofs << std::endl;
      std::cout << ";" << std::endl;
      //std::cout << std::flush;
    }

    std::cout << "KindHost - finished" << std::endl;
    lsh_.clear();
    //std::cout << std::flush;

    return true;
  }

  bool load_lsh_mother(std::vector<FeatureVector>& mother,
		       std::vector<uint>& len, 
		       const std::vector<std::string>& files) const
  {
    for (uint i=0; i!=files.size(); ++i) {
      Glob glob(files[i].c_str());
      if (glob.empty()) {
	std::ostringstream os;
	os << files[i] << ": no matches found";
	throw os.str().c_str();
	//return false;
      }
      Glob::const_iterator p;
      for (p=glob.begin(); p!=glob.end(); ++p) {
	double elapsed = 0.0;
	typename LDF::Loader* loader=ldf_.get_loader(p->c_str());
	if (loader==NULL) return false;
#ifdef HAVE_MPI
	if (MPI::COMM_WORLD.Get_rank()==0) {
#endif
	  std::cout << "loading " << *p << std::flush;
#ifdef HAVE_MPI
	}
#endif
	while (true) {
	  boost::timer tm;
	  Data* d = loader->get();
	  elapsed += tm.elapsed();
	  if (d==NULL) break;
	  mother.insert(mother.end(),d->feat.begin(),d->feat.end());
	  len.push_back(d->size());
	  delete d;
	}
#ifdef HAVE_MPI
	if (MPI::COMM_WORLD.Get_rank()==0) {
#endif
	  std::cout << " (" << elapsed << "s) done." << std::endl;
#ifdef HAVE_MPI
	}
#endif
	delete loader;
      }
    }
    return true;
  }

  uint load_lsh_data(std::vector<HashData>& data,
		    const std::vector<uint>& len)
  {
    uint sid=0, pos=0, off=0;
    uint cnt=0;

    while (true) {
      uint win = len[sid];
      uint slide = len[sid];

      while (pos+win > len[sid]) {
	off += len[sid] - pos;
	pos = 0;
	++sid;
	if (sid==len.size()) return cnt;
	win = len[sid];
	slide = len[sid];
      }

      if (opts_.verbose)
	std::cout << "storing " << "(" 
		  << sid << "," << pos << "," 
		  << win << "," << off << ")" 
		  << std::endl << std::flush; 

      HashData hd(sid, pos, win, off);
      data.push_back(hd);
      ++cnt;

      off += slide;
      pos += slide;
    }

    return 0;
  }

private:
  LSH& lsh_;
  const LDF& ldf_;
  const KNL& knl_;
  const Options& opts_;
};

#ifdef HAVE_MPI
class MPIState
{
public:
  MPIState(int& argc, char**& argv)
  {
    MPI::Init(argc, const_cast<char**&>(argv));
  }

  ~MPIState()
  {
    if (MPI::Is_initialized())
      MPI::Finalize();
  }
};
#endif

#endif	
