#include <cmath>
#include <boost/thread.hpp>

#include "KernelCalculator.h"

using namespace std;
using namespace boost;

template <class K>
class CalcMatrix
{
  int n_th_;
  int th_no_;
  const vector<FeatureVector>& ex_;
  const K& kernel_;
  bool rescale_;
  KernelMatrix& km_;

public:
  CalcMatrix(int n_th, int th_no, 
	     const vector<FeatureVector>& ex, const K& kernel, 
	     bool rescale, KernelMatrix& km)
    : n_th_(n_th), th_no_(th_no), 
      ex_(ex), kernel_(kernel), 
      rescale_(rescale), km_(km) {}

  void operator()()
  {
    int cnt=0;
    int n = ex_.size();
    for (int i=0; i!=n; ++i) {
      for (int j=i; j!=n; ++j) {
	if (cnt++%n_th_==th_no_) {
	  km_(i,j) = kernel_(ex_[i], ex_[j], rescale_, km_.scl(i,j));
	  if (i!=j) { 
	    km_(j,i)=km_(i,j);
	    km_.scl(j,i)=km_.scl(i,j);
	  }
	}
      }
    }
  }
};

template <class K>
void KernelMatrix::
calculate(const vector<FeatureVector>& ex, const K& kernel, 
	  bool logarithm, bool normalize, bool rescale, int n_th)
{
  typedef CalcMatrix<K> CalcMatrix;
  resize(ex.size());

  if (n_th>1) {
    vector<thread*> th(n_th);
    vector<CalcMatrix*> calc(n_th);
    for (int t=0; t!=n_th; ++t) {
      calc[t] = new CalcMatrix(n_th, t, ex, kernel, rescale, *this);
      th[t] = new boost::thread(*calc[t]);
    }
    for (int t=0; t!=n_th; ++t) {
      th[t]->join();
      delete calc[t];
      delete th[t];
    }
  } 
  else {
    CalcMatrix calc(1, 0, ex, kernel, rescale, *this);
    calc();
  }
  for (int i=0; i!=n_; ++i) {
    diag_score_[i] = score_[i][i];
    diag_scale_[i] = scale_[i][i];
  }

  if (logarithm) {
    for (int i=0; i!=n_; ++i) {
      for (int j=i; j!=n_; ++j) {
	score_[i][j] = log(score_[i][j]);
	scale_[i][j] = log(scale_[i][j]);
	if (i!=j) { 
	  score_[j][i] = score_[i][j];
	  scale_[j][i] = scale_[i][j];
	}
      }
    }
    for (int i=0; i!=n_; ++i) {
      diag_score_[i] = score_[i][i];
      diag_scale_[i] = scale_[i][i];
    }

    if (normalize) { 
      if (rescale) { 
	// logarithm && normalize && rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] -= scale_[i][j];
	    score_[i][j] /= sqrt( (diag_score_[i]-diag_scale_[i]) * 
				  (diag_score_[j]-diag_scale_[j]) );
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
      else {
	// logarithm && normalize && !rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] /= sqrt(diag_score_[i] * diag_score_[j]);
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
    }
    else {
      if (rescale) { 
	// logarithm && !normalize && rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] -= scale_[i][j];
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
      else {
	// logarithm && !normalize && !rescale
      }
    }
  }
  else {
    if (normalize) { 
      if (rescale) { 
	// !logarithm && normalize && rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] /= sqrt(diag_score_[i] * diag_score_[j]);
	    score_[i][j] /= scale_[i][j] / sqrt(diag_scale_[i] * diag_scale_[j]);
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
      else {
	// !logarithm && normalize && !rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] /= sqrt(diag_score_[i] * diag_score_[j]);
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
    }
    else {
      if (rescale) { 
	// !logarithm && !normalize && rescale
	for (int i=0; i!=n_; ++i) {
	  for (int j=i; j!=n_; ++j) {
	    score_[i][j] /= scale_[i][j];
	    if (i!=j) { 
	      score_[j][i] = score_[i][j];
	    }
	  }
	}
      }
      else {
	// !logarithm && !normalize && !rescale
      }
    }
  }
}

void KernelMatrix::
print()
{
  cout << "Kernel Matrix:" << endl << flush;
  for (int i=0; i!=n_; ++i) {
    for (int j=0; j!=n_; ++j) {
      cout << score_[i][j] << " ";      
    }
    cout << endl << flush;
  }

  cout << "Kernel Matrix Scale:" << endl << flush;
  for (int i=0; i!=n_; ++i) {
    for (int j=0; j!=n_; ++j) {
      cout << scale_[i][j] << " ";      
    }
    cout << endl << flush;
  }
}

template <class K>
class PrepSearch
{
  int n_th_;
  int th_no_;
  const vector<FeatureVector>& ex_;
  const K& kernel_;
  bool rescale_;
  vector<double>& ex_score_;
  vector<double>& ex_scale_;

public:
  PrepSearch(int n_th, int th_no, 
	     const vector<FeatureVector>& ex, const K& kernel, 
	     bool rescale, vector<double>& ex_score, vector<double>& ex_scale)
    : n_th_(n_th), th_no_(th_no), 
      ex_(ex), kernel_(kernel), 
      rescale_(rescale), ex_score_(ex_score), ex_scale_(ex_scale) {}
  
  void operator()()
  {
    int cnt=0;
    int n = ex_.size();
    for (int i=0; i!=n; ++i) {
      if (cnt++%n_th_==th_no_) {
	ex_score_[i] = kernel_(ex_[i], ex_[i], rescale_, ex_scale_[i]);
      }
    }
  }
};

template <class K>
void KernelSearch::
prepare(const vector<FeatureVector>& db, const K& kernel, 
	bool rescale, int n_th)
{
  typedef PrepSearch<K> PrepSearch;
  resize(db.size());

  if (n_th>1) {
    vector<thread*> th(n_th);
    vector<PrepSearch*> calc(n_th);
    for (int t=0; t!=n_th; ++t) {
      calc[t] = new PrepSearch(n_th, t, db, kernel, rescale, db_score_, db_scale_);
      th[t] = new boost::thread(*calc[t]);
    }
    for (int t=0; t!=n_th; ++t) {
      th[t]->join();
      delete calc[t];
      delete th[t];
    }
  } 
  else {
    PrepSearch calc(1, 0, db, kernel, rescale, db_score_, db_scale_);
    calc();
  }

}

template <class K>
class CalcSearch
{
  int n_th_;
  int th_no_;
  const vector<FeatureVector>& qu_;
  const vector<FeatureVector>& db_;
  const K& kernel_;
  bool rescale_;
  KernelSearch& ks_;

public:
  CalcSearch(int n_th, int th_no, 
	     const vector<FeatureVector>& qu, 
	     const vector<FeatureVector>& db, const K& kernel, 
	     bool rescale, KernelSearch& ks)
    : n_th_(n_th), th_no_(th_no), 
      qu_(qu), db_(db), kernel_(kernel), 
      rescale_(rescale), ks_(ks) {}
  
  void operator()()
  {
    int cnt=0;
    int nqu = qu_.size();
    int ndb = db_.size();
    for (int i=0; i!=nqu; ++i) {
      for (int j=0; j!=ndb; ++j) {
	if (cnt++%n_th_==th_no_) {
	  ks_(i,j) = kernel_(qu_[i], db_[j], rescale_, ks_.scl(i,j));
	}
      }
    }
  }
};

template <class K>
void KernelSearch::
calculate(const vector<FeatureVector>& qu,
	  const vector<FeatureVector>& db, const K& kernel, 
	  bool logarithm, bool normalize, bool rescale, int n_th)
{
  typedef PrepSearch<K> PrepSearch;
  typedef CalcSearch<K> CalcSearch;
  resize(qu.size(), db.size());

  if (n_th>1) {
    vector<thread*> th(n_th);
    vector<PrepSearch*> calc(n_th);
    for (int t=0; t!=n_th; ++t) {
      calc[t] = new PrepSearch(n_th, t, qu, kernel, rescale, qu_score_, qu_scale_);
      th[t] = new boost::thread(*calc[t]);
    }
    for (int t=0; t!=n_th; ++t) {
      th[t]->join();
      delete calc[t];
      delete th[t];
    }
  } 
  else {
    PrepSearch calc(1, 0, qu, kernel, rescale, qu_score_, qu_scale_);
    calc();
  }

  if (n_th>1) {
    vector<thread*> th(n_th);
    vector<CalcSearch*> calc(n_th);
    for (int t=0; t!=n_th; ++t) {
      calc[t] = new CalcSearch(n_th, t, qu, db, kernel, rescale, *this);
      th[t] = new boost::thread(*calc[t]);
    }
    for (int t=0; t!=n_th; ++t) {
      th[t]->join();
      delete calc[t];
      delete th[t];
    }
  } 
  else {
    CalcSearch calc(1, 0, qu, db, kernel, rescale, *this);
    calc();
  }

  if (logarithm) {
    for (int i=0; i!=nqu_; ++i) {
      for (int j=0; j!=ndb_; ++j) {
	score_[i][j] = log(score_[i][j]);
	scale_[i][j] = log(scale_[i][j]);
      }
    }
    for (int i=0; i!=nqu_; ++i) {
      qu_score_[i] = log(qu_score_[i]);
      qu_scale_[i] = log(qu_scale_[i]);
    }
    for (int i=0; i!=ndb_; ++i) {
      db_score_[i] = log(db_score_[i]);
      db_scale_[i] = log(db_scale_[i]);
    }

    if (normalize) { 
      if (rescale) { 
	// logarithm && normalize && rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] -= scale_[i][j];
	    score_[i][j] /= sqrt( (qu_score_[i]-qu_scale_[i]) * 
				  (db_score_[j]-db_scale_[j]) );
	  }
	}
      }
      else {
	// logarithm && normalize && !rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] /= sqrt(qu_score_[i] * db_score_[j]);
	  }
	}
      }
    }
    else {
      if (rescale) { 
	// logarithm && !normalize && rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] -= scale_[i][j];
	  }
	}
      }
      else {
	// logarithm && !normalize && !rescale
      }
    }
  }
  else {
    if (normalize) { 
      if (rescale) { 
	// !logarithm && normalize && rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] /= sqrt(qu_score_[i] * db_score_[j]);
	    score_[i][j] /= scale_[i][j] / sqrt(qu_scale_[i] * db_scale_[j]);
	  }
	}
      }
      else {
	// !logarithm && normalize && !rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] /= sqrt(qu_score_[i] * db_score_[j]);
	  }
	}
      }
    }
    else {
      if (rescale) { 
	// !logarithm && !normalize && rescale
	for (int i=0; i!=nqu_; ++i) {
	  for (int j=0; j!=ndb_; ++j) {
	    score_[i][j] /= scale_[i][j];
	  }
	}
      }
      else {
	// !logarithm && !normalize && !rescale
      }
    }
  }
}

void KernelSearch::
print()
{
  cout << "Kernel Search:" << endl << flush;
  for (int i=0; i!=nqu_; ++i) {
    for (int j=0; j!=ndb_; ++j) {
      cout << score_[i][j] << " ";      
    }
    cout << endl << flush;
  }

  cout << "Kernel Search Scale:" << endl << flush;
  for (int i=0; i!=nqu_; ++i) {
    for (int j=0; j!=ndb_; ++j) {
      cout << scale_[i][j] << " ";      
    }
    cout << endl << flush;
  }
}

// instantiation

#include "BPLAKernel.h"

template void KernelMatrix::calculate<BPLAKernel> 
(const vector<FeatureVector>&, const BPLAKernel&, bool, bool, bool, int);

template void KernelSearch::prepare<BPLAKernel> 
(const vector<FeatureVector>&, const BPLAKernel&, bool, int);

template void KernelSearch::calculate<BPLAKernel> 
(const vector<FeatureVector>&, const vector<FeatureVector>&, 
 const BPLAKernel&, bool, bool, bool, int);
