#ifndef __INC_KERNEL_CALCULATOR_H__
#define __INC_KERNEL_CALCULATOR_H__

#include <vector>
#include <cassert>
#include <boost/multi_array.hpp>

#include "Utility.h"

class KernelMatrix {
private:
  boost::multi_array<double,2> score_;
  boost::multi_array<double,2> scale_;
  std::vector<double> diag_score_;
  std::vector<double> diag_scale_;
  int n_;

public:
  KernelMatrix() : score_(), scale_(), diag_score_(), diag_scale_() {}
  
  void resize(int n)
  {
    score_.resize(boost::extents[n][n]);
    scale_.resize(boost::extents[n][n]);
    diag_score_.resize(n);
    diag_scale_.resize(n);
    n_ = n;
  }

  double& operator()(int i, int j)
  {
    return score_[i][j];
  }
  const double& operator()(int i, int j) const
  {
    return score_[i][j];
  }

  double& scl(int i, int j)
  {
    return scale_[i][j];
  }
  const double& scl(int i, int j) const
  {
    return scale_[i][j];
  }

  void get_diag(std::vector<double>& diag_score, 
		std::vector<double>& diag_scale) const
  {
    diag_score = diag_score_;
    diag_scale = diag_scale_;
  }
  void set_diag(const std::vector<double>& diag_score, 
		const std::vector<double>& diag_scale) 
  {
    diag_score_ = diag_score;
    diag_scale_ = diag_scale;
  }

  template <class K>
  void calculate(const std::vector<FeatureVector>& ex, const K& kernel, 
		 bool logarithm, bool normalize, bool rescale, int n_th);

  void print();
};

class KernelSearch {
private:
  boost::multi_array<double,2> score_;
  boost::multi_array<double,2> scale_;
  std::vector<double> qu_score_;
  std::vector<double> qu_scale_;
  std::vector<double> db_score_;
  std::vector<double> db_scale_;
  int nqu_;
  int ndb_;

public:
  KernelSearch() : score_(), scale_(), 
		   qu_score_(), qu_scale_(), db_score_(), db_scale_() {}

  void resize(int ndb)
  {
    db_score_.resize(ndb);
    db_scale_.resize(ndb);
    ndb_ = ndb;
  }

  void resize(int nqu, int ndb)
  {
    assert(db_score_.size()==ndb && 
	   db_scale_.size()==ndb &&
	   ndb_==ndb);

    score_.resize(boost::extents[nqu][ndb]);
    scale_.resize(boost::extents[nqu][ndb]);
    qu_score_.resize(nqu);
    qu_scale_.resize(nqu);
    nqu_ = nqu;
  }

  double& operator()(int j)
  {
    return score_[0][j];
  }
  const double& operator()(int j) const
  {
    return score_[0][j];
  }
  double& operator()(int i, int j)
  {
    return score_[i][j];
  }
  const double& operator()(int i, int j) const
  {
    return score_[i][j];
  }

  double& scl(int j)
  {
    return scale_[0][j];
  }
  const double& scl(int j) const
  {
    return scale_[0][j];
  }
  double& scl(int i, int j)
  {
    return scale_[i][j];
  }
  const double& scl(int i, int j) const
  {
    return scale_[i][j];
  }

  void get_db(std::vector<double>& db_score, 
	      std::vector<double>& db_scale) const
  {
    db_score = db_score_;
    db_scale = db_scale_;
  }
  void set_db(const std::vector<double>& db_score, 
	      const std::vector<double>& db_scale) 
  {
    assert(db_score.size()==db_scale.size());

    db_score_ = db_score;
    db_scale_ = db_scale;
    ndb_ = db_score.size();
  }

  template <class K>
  void prepare(const std::vector<FeatureVector>& db, const K& kernel, 
	       bool rescale, int n_th);  

  template <class K>
  void calculate(const FeatureVector& q, 
		 const std::vector<FeatureVector>& db, const K& kernel, 
		 bool logarithm, bool normalize, bool rescale, int n_th)
  {
    std::vector<FeatureVector> qu(1,q);
    calculate(qu,db,kernel,logarithm,normalize,rescale,n_th);
  }

  template <class K>
  void calculate(const std::vector<FeatureVector>& qu, 
		 const std::vector<FeatureVector>& db, const K& kernel, 
		 bool logarithm, bool normalize, bool rescale, int n_th); 

  void print();
};

#endif 
