#include <list>
#include <deque>
#include <string>
#include <algorithm>
#include <numeric>
#include <cassert>
#include <cmath>
#include <boost/dynamic_bitset.hpp>
#include <boost/random.hpp>
#include <boost/math/distributions.hpp>

#include "KernelizedLSH.h"
#include "KernelCalculator.h"

using namespace std;
using namespace boost;

#define THSH_SIZE 20
#define LOSS_SIZE 10
#define MAX_NPROB 200
const double thsh_[THSH_SIZE] = {0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40, 0.45, 
				 0.50, 0.55, 0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95};
const double loss_[LOSS_SIZE] = {0.01, 0.03, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40, 0.50, 0.60}; 

//---------- Kernelized LSH for Implicit Feature Space -----------------
//---------- Kulis and Grauman, 2009
             
template <class K>
KernelizedLSH<K>::
KernelizedLSH(const LSH::Options& opts, 
	      uint ntrn, uint nsmp, bool rescale, const K& kernel)
  : LSH(opts,false), 
    ntrn_(ntrn), nsmp_(nsmp), rescale_(rescale), kernel_(kernel)
{
  assert(nsec_ == 2);
  assert(nsmp_ <= ntrn_);
}

// generate hashing functions using implicit random hyperplanes
template <class K>
vector<typename KernelizedLSH<K>::HashFunction> KernelizedLSH<K>::
hash_function(const std::vector<std::vector<double> >& trn_matrix, 
	      uint ntrn, uint nsmp, uint ntab, uint nand, 
	      RandomGenerator& rng)
{
  assert(trn_matrix.size() == ntrn);
  for (uint i=0; i!=ntrn; ++i)
    assert(trn_matrix[i].size() == ntrn);
    
  vector<HashFunction> hf;

  vector<vector<double> > sqrt_km = sqrt_symmetric_matrix(trn_matrix);

  /* test K^(1/2)^2 == K
  vector<vector<double> > sqrt_km2(ntrn,vector<double>(ntrn));
  for (int i=0; i!=ntrn; ++i) 
    for (int j=0; j!=ntrn; ++j) 
      sqrt_km2[i][j] = 0.0;
  for (int i=0; i!=ntrn; ++i) 
    for (int j=0; j!=ntrn; ++j) 
      for (int k=0; k!=ntrn; ++k) 
	sqrt_km2[i][j] += sqrt_km[i][k] * sqrt_km[k][j];
  for (int i=0; i!=ntrn; ++i) {
    for (int j=0; j!=ntrn; ++j) {
      std::cout << sqrt_km[i][j] << "," << sqrt_km2[i][j] << "," 
		<< trn_matrix[i][j] << " ";
    }
    std::cout << "<<" << std::endl;
  } 
  */

  hf.resize(ntab);
  for (uint i=0; i!=ntab; ++i) {
    hf[i].resize(nand);
    for (uint j=0; j!=nand; ++j) {
      vector<double> ed(ntrn,-1.0/ntrn);
      RandomShuffler rns(rng);

      for (uint k=0; k!=nsmp; ++k) 
	ed[k] += 1.0/nsmp;
      random_shuffle(ed.begin(),ed.end(),rns);

      hf[i][j].resize(ntrn);
      for (uint k=0; k!=ntrn; ++k) 
	hf[i][j][k] = inner_product(ed.begin(),ed.end(),sqrt_km[k].begin(),0.0);
    }
  }

  return hf;
}

template <class K>
void KernelizedLSH<K>::
make_function(uint nthr)
{
  assert(ntrn_ <= data_.size());

  trn_.resize(ntrn_);

  deque<bool> pick(data_.size(),false); // do not use vector<bool> broken in STL
  RandomShuffler rns(rng_);
  fill_n(pick.begin(),ntrn_,true);
  random_shuffle(pick.begin(),pick.end(),rns);

  for (uint p=0, i=0; i!=data_.size(); ++i) 
    if(pick[i]) trn_[p++] = get_feat(i);

  /* selection with replacements
  uniform_int<> unif_dist(0,data_.size()-1);
  variate_generator<RandomGenerator&,uniform_int<> > unif_samp(rng_,unif_dist);

  for (uint i=0; i!=ntrn_; ++i) 
    trn_[i] = get_feat(unif_samp());
  */

  KernelSearch sh;
  sh.prepare(trn_,kernel_,rescale_,nthr);
  sh.get_db(trn_score_,trn_scale_);

  KernelMatrix mx;
  mx.calculate(trn_,kernel_,false,normalize_,rescale_,nthr);
  trn_matrix_.resize(ntrn_,vector<double>(ntrn_));
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      trn_matrix_[i][j] = mx(i,j);

  hf_ = hash_function(trn_matrix_,ntrn_,nsmp_,ntab_,nand_,rng_);
}

template <class K>
void KernelizedLSH<K>::
make_index(uint nthr)
{
  KernelSearch sh;
  sh.set_db(trn_score_,trn_scale_);

  for (uint i=0; i!=data_.size(); ++i) {
    sh.calculate(get_feat(i),trn_,kernel_,false,normalize_,rescale_,nthr);
    //sh.print();
    vector<double> ks(ntrn_);
    for (uint j=0; j!=ntrn_; ++j) 
      ks[j] = sh(j);

    for (uint j=0; j!=ntab_; ++j) {
      dynamic_bitset<> b(nand_); // initialized by 0s
      for (uint k=0; k!=nand_; ++k) {
	double g = inner_product(ks.begin(),ks.end(),hf_[j][k].begin(),0.0);
	if(g > 0) b.set(k,1);
      }
      /*
      std::cout << "hashed to " << j << " " 
		<< b << "(" << b.to_ulong() << ")" 
		<< std::endl << std::flush;
      */
      table_[j][b.to_ulong()].push_back(i);
    }
  }
}

template <class K>
void KernelizedLSH<K>::
print_function() 
{
  for (uint i=0; i!=ntab_; ++i) {
    cout << "LSH function " << i << ":" << endl << flush;
    for (uint j=0; j!=nand_; ++j) {
      cout << "hyper plane " << j << ":" << endl << flush;
      for (uint k=0; k!=ntrn_; ++k) {
	cout << hf_[i][j][k] << " ";
      }
      cout << endl << flush;
    }
  }
}

template <class K>
void KernelizedLSH<K>::
print_index() 
{
  cout << " ntab:" << ntab_ << " nand:" << nand_ << " nsec:" << nsec_ 
       << " normalize:" << normalize_ << " rescale:" << rescale_
       << endl << flush;
  cout << "ntrn:" << ntrn_ << " nsmp:" << nsmp_
       << endl << flush;
  
  cout << "thsh: " << endl << flush;
  for (uint i=0; i!=THSH_SIZE; ++i) 
    cout << thsh_[i] << " ";
  cout << endl << flush;
  
  cout << "loss: " << endl << flush;
  for (uint i=0; i!=LOSS_SIZE; ++i) 
    cout << loss_[i] << " ";
  cout << endl << flush;
  
  cout << "opt_nprb: " << endl << flush;
  for (uint i=0; i!=THSH_SIZE; ++i) {
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      cout << opt_nprb_[i][j] << " ";
    cout << endl << flush;
  }
  
  for (uint i=0; i!=ntab_; ++i) {
    cout << "LSH table " << i << ":" << endl << flush;
    for (uint j=0; j!=table_[i].size(); ++j) {
      if (table_[i][j].empty()) continue;
      cout << "BACKET " << j << ":" << endl << flush;
      HashBacket::iterator k = table_[i][j].begin();
      HashBacket::iterator ke = table_[i][j].end();
      for ( ; k!=ke; ++k) {
	const HashData& d = get_data(*k);

	cout << "(" 
	     << d.sid_ << "," << d.pos_ << "," 
	     << d.win_ << "," << d.off_ << ")" 
	     << endl << flush; 
      }
    }
  }
}

template <class K>
void KernelizedLSH<K>::
optimize_nprb(uint nthr)
{
  // resample virtual queries
  vector<FeatureVector> vq(ntrn_);
  vector<vector<double> > vq_ks(ntrn_,vector<double>(ntrn_));
  KernelSearch vq_sh;

  deque<bool> pick(data_.size(),false); // do not use vector<bool> broken in STL
  RandomShuffler rns(rng_);

  fill_n(pick.begin(),ntrn_,true);
  random_shuffle(pick.begin(),pick.end(),rns);

  for (uint p=0, i=0; i!=data_.size(); ++i) 
    if (pick[i]) vq[p++] = get_feat(i);

  vq_sh.prepare(trn_,kernel_,rescale_,nthr);
  vq_sh.calculate(vq,trn_,kernel_,false,normalize_,rescale_,nthr);
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      vq_ks[i][j] = vq_sh(i,j);

  // edge[i][j] = k means that 
  // trn_[i] captures data_[j] only by the threshold < thsh_[k] 
  vector<vector<double> > kv(ntrn_,vector<double>(data_.size()));
  vector<vector<uint> > edge(ntrn_,vector<uint>(data_.size(),THSH_SIZE));

  KernelSearch qu_sh;
  vector<double> score;
  vector<double> scale;
  typename K::Options qu_opts(K::QUERY);
  K qu_kernel(qu_opts);
  bool qu_logarithm = true;
  bool qu_rescale = true;
  qu_sh.prepare(vq,qu_kernel,qu_rescale,nthr);
  qu_sh.get_db(score,scale);

  for (uint i=0; i!=data_.size(); ++i) {
    KernelSearch sh;
    sh.set_db(score,scale); 
    sh.calculate(get_feat(i),vq,qu_kernel,qu_logarithm,true,qu_rescale,nthr);
    for (uint j=0; j!=ntrn_; ++j)
      kv[j][i] = sh(j);
  }
  
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=data_.size(); ++j) 
      for (uint k=0; k!=THSH_SIZE; ++k) 
	if (kv[i][j] < thsh_[k]) {
	  edge[i][j] = k;
	  break;
	}

  // nhit[i][j] = f means that
  // for i+1 multi-probe and thsh_[j], fraction f hits are missed
  vector<uint> nexact(THSH_SIZE,0);
  vector<vector<double> > fmiss(MAX_NPROB,vector<double>(THSH_SIZE,0.0));

  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      for (uint k=0; k!=data_.size(); ++k) 
	if (i < edge[j][k]) nexact[i]++;

  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=ntrn_; ++j) {
      vector<HashHit> hit;
      probe(hit,vq_ks[j],i+1);
      for (uint k=0; k!=hit.size(); ++k)
	for (uint l=0; l!=edge[j][hit[k].i]; ++l)
	  fmiss[i][l] += 1.0;
    }
  }
  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=THSH_SIZE; ++j) {
      if (nexact[j]==0) fmiss[i][j] = 0.0;
      else fmiss[i][j] = 1.0 - fmiss[i][j] / nexact[j];
    }
  }

  // opt_probe_[i][j] = k means that 
  // for thsh_[i] and loss_[j], k is the optimized nprb 
  opt_nprb_.assign(THSH_SIZE,vector<uint>(LOSS_SIZE,MAX_NPROB));
  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      for (uint k=0; k!=MAX_NPROB; ++k) 
	if (fmiss[k][i] < loss_[j]) {
	  opt_nprb_[i][j] = k+1;
	  break;
	}

  // use trn_ as virtual queries
  /* 
  // edge[i][j] = k means that 
  // trn_[i] captures data_[j] only by the threshold < thsh_[k] 
  vector<vector<double> > kv(ntrn_,vector<double>(data_.size()));
  vector<vector<uint> > edge(ntrn_,vector<uint>(data_.size(),THSH_SIZE));

  KernelSearch qu_sh;
  vector<double> score;
  vector<double> scale;
  typename K::Options qu_opts(K::QUERY);
  K qu_kernel(qu_opts);
  bool qu_logarithm = true;
  bool qu_rescale = true;
  qu_sh.prepare(trn_,qu_kernel,qu_rescale,nthr);
  qu_sh.get_db(score,scale);

  for (uint i=0; i!=data_.size(); ++i) {
    KernelSearch sh;
    sh.set_db(score,scale); 
    sh.calculate(get_feat(i),trn_,qu_kernel,qu_logarithm,true,qu_rescale,nthr);
    for (uint j=0; j!=ntrn_; ++j)
      kv[j][i] = sh(j);
  }
  
  for (uint i=0; i!=ntrn_; ++i) 
    for (uint j=0; j!=data_.size(); ++j) 
      for (uint k=0; k!=THSH_SIZE; ++k) 
	if (kv[i][j] < thsh_[k]) {
	  edge[i][j] = k;
	  break;
	}

  // nhit[i][j] = f means that
  // for i+1 multi-probe and thsh_[j], fraction f hits are missed
  vector<uint> nexact(THSH_SIZE,0);
  vector<vector<double> > fmiss(MAX_NPROB,vector<double>(THSH_SIZE,0.0));

  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=ntrn_; ++j) 
      for (uint k=0; k!=data_.size(); ++k) 
	if (i < edge[j][k]) nexact[i]++;

  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=ntrn_; ++j) {
      vector<HashHit> hit;
      probe(hit,trn_matrix_[j],i+1);
      for (uint k=0; k!=hit.size(); ++k)
	for (uint l=0; l!=edge[j][hit[k].i]; ++l)
	  fmiss[i][l] += 1.0;
    }
  }
  for (uint i=0; i!=MAX_NPROB; ++i) {
    for (uint j=0; j!=THSH_SIZE; ++j) {
      if (nexact[j]==0) fmiss[i][j] = 0.0;
      else fmiss[i][j] = 1.0 - fmiss[i][j] / nexact[j];
    }
  }

  // opt_probe_[i][j] = k means that 
  // for thsh_[i] and loss_[j], k is the optimized nprb 
  opt_nprb_.assign(THSH_SIZE,vector<uint>(LOSS_SIZE,MAX_NPROB));
  for (uint i=0; i!=THSH_SIZE; ++i) 
    for (uint j=0; j!=LOSS_SIZE; ++j) 
      for (uint k=0; k!=MAX_NPROB; ++k) 
	if (fmiss[k][i] < loss_[j]) {
	  opt_nprb_[i][j] = k+1;
	  break;
	}
  */
}

template <class K>
uint KernelizedLSH<K>::
estimate_nprb(double thsh, double loss) {
  uint i, j;

  for (i=0; i!=THSH_SIZE; ++i) 
    if (thsh < thsh_[i]) break;
  if (i==0) return MAX_NPROB;

  for (j=0; j!=LOSS_SIZE; ++j) 
    if (loss < loss_[j]) break;
  if (j==0) return MAX_NPROB;

  return opt_nprb_[i-1][j-1];
}

template <class K>
void KernelizedLSH<K>::
probe(vector<HashHit>& hit, const FeatureVector& x,
      uint nprb, uint nthr)
{
  vector<double> ks(ntrn_);
  KernelSearch sh;
  sh.set_db(trn_score_,trn_scale_);
  sh.calculate(x,trn_,kernel_,false,normalize_,rescale_,nthr);
  for (uint i=0; i!=ntrn_; ++i)
    ks[i] = sh(i);

  probe(hit,ks,nprb);
}

template <class K>
void KernelizedLSH<K>::
probe(vector<HashHit>& hit, const vector<double>& ks, 
      uint nprb)
{
  assert(nprb <= dblpow(2,nand_));

  list<HashHit> tmp; // use list for massive push_back()

  for (uint i=0; i!=ntab_; ++i) {
    vector<HashValue> seq;
    multi_probe(seq,ks,i,nprb);
    for (uint j=0; j!=seq.size(); ++j) {
      HashBacket::iterator k = table_[i][seq[j]].begin();
      HashBacket::iterator ke = table_[i][seq[j]].end();
      for ( ; k!=ke; ++k) {
	HashHit h;
	h.i = *k;
	h.s = 0.0;
	tmp.push_back(h); 
      }
    }
  }

  hit.assign(tmp.begin(),tmp.end());
  sort(hit.begin(),hit.end(),HashHitLess());
  vector<HashHit>::iterator he = unique(hit.begin(),hit.end(),HashHitSame());
  hit.erase(he,hit.end());
}


template <class K>
void KernelizedLSH<K>::
multi_probe(vector<HashValue>& seq, 
	    const vector<double>& ks, uint idx, uint nprb)
{
  dynamic_bitset<> b(nand_); // initialized by 0s
  vector<HashMargin> margin(nand_);

  for (uint i=0; i!=nand_; ++i) {
    double g = inner_product(ks.begin(),ks.end(),hf_[idx][i].begin(),0.0);
    if(g > 0) b.set(i,1);
    margin[i].first = i;
    margin[i].second = fabs(g);
  }
  sort(margin.begin(),margin.end(),HashMarginLess());
  // now, margin[i].first represents the index of the hyperplane 
  // that gives the i-th smallest margin

  // set the actual i-th smallest margin
  vector<double> z(nand_);
  for (uint i=0; i!=nand_; ++i) 
    z[i] = margin[i].second;

  // true in the i-th element of HashProbe means 
  // that we probe the neighboring backet of the i-th smallest margin.
  // of course we first probe the hash backet itself.
  vector<HashProbe> hp(nprb);
  hp[0] = HashProbe(nand_,false);

  list<ProbeCandidate> heap;
  HashProbe p(nand_,false); 
  p[0] = true;
  heap.push_back(ProbeCandidate(p,z[0]));

  for (uint i=1; i!=nprb; ++i) {
    list<ProbeCandidate>::iterator hi = 
      min_element(heap.begin(),heap.end(),ProbeCandidateLess());
    hp[i] = (*hi).first;

    extend_heap(heap,hi,z,nand_);
  }

  seq.resize(nprb);
  for (uint i=0; i!=nprb; ++i) {
    // for the ease of computation, we map the order of margins
    // into the order of hyperplanes 
    dynamic_bitset<> pt(b); 
    for (uint j=0; j!=nand_; ++j) {
      if (hp[i][j]) pt.flip(margin[j].first);
    }
    seq[i] = pt.to_ulong();
  }
}

// instantiation

#include "BPLAKernel.h"
template class KernelizedLSH<BPLAKernel>;
