#ifndef __INC_KERNELIZED_LSH_H__
#define __INC_KERNELIZED_LSH_H__

#include <vector>

#include <limits.h>  // necessary for boost/serialization
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>

#include "Utility.h"
#include "LSH.h"

//---------- Kernelized LSH for Implicit Feature Space -----------------
//---------- Kulis and Grauman, 2009

template <class K>
class KernelizedLSH : public LSH
{
public:
  typedef std::vector<WeightVector> HashFunction;

public:
  KernelizedLSH(const LSH::Options& opts, 
		uint ntrn, uint nsmp, bool rescale, const K& kernel);
  KernelizedLSH() : LSH(), ntrn_(0), nsmp_(0), rescale_(false), kernel_() {}
  ~KernelizedLSH() {}

public:
  void make_function(uint nthr);
  void make_index(uint nthr);
  void print_function();
  void print_index();
  void probe(std::vector<HashHit>& hit, const FeatureVector& x, 
	     uint nprb, uint nthr);
  void probe(std::vector<HashHit>& hit, const std::vector<double>& ks, 
	     uint nprb);
  void optimize_nprb(uint nthr);
  uint estimate_nprb(double thsh, double loss);

  static std::vector<HashFunction> 
  hash_function(const std::vector<std::vector<double> >& trn_matrix, 
		uint ntrn, uint nsmp, uint ntab, uint nand, 
		RandomGenerator& rng);

private:
  void multi_probe(std::vector<HashValue>& seq, 
		   const std::vector<double>& ks, uint idx, uint nprb);

  // pseudo override for pure virtual functions in class LSH
  // these functions are unnecessary or overloaded in this child class
  HashValue hash_value(const FeatureVector& x, uint idx) { return 0; } 
  void multi_probe(std::vector<HashValue>& seq, 
		   const FeatureVector& x, uint idx, uint nprb) {}

private:
  std::vector<HashFunction> hf_;
  std::vector<FeatureVector> trn_;
  std::vector<double> trn_score_;
  std::vector<double> trn_scale_;
  std::vector<std::vector<double> > trn_matrix_;
  std::vector<std::vector<uint> > opt_nprb_; 
  const uint ntrn_;
  const uint nsmp_;
  const bool rescale_;
  const K kernel_;

private:
  friend class boost::serialization::access;
  template <class Archive> 
  void serialize(Archive& ar, const uint ver)
  {
    ar & boost::serialization::base_object<LSH>(*this);
    ar & hf_ & trn_ & trn_score_ & trn_scale_ & opt_nprb_;
    ar & const_cast<uint&>(ntrn_) & const_cast<uint&>(nsmp_);
    ar & const_cast<bool&>(rescale_) & const_cast<K&>(kernel_);
  }
};

#endif
