#include <algorithm>
#include <cassert>
#include <boost/random.hpp>
#include <boost/math/distributions.hpp>

#include "LSH.h"

using namespace std;
using namespace boost;
namespace po = boost::program_options;

//---------- Base Class for Various LSH Schemes -------------------------
//---------- Indyk and Motwani, 1998

void LSH::Options::
add_options(po::options_description& desc)
{
  desc.add_options()
    ("table,T", 
     po::value<uint>(&ntab)->default_value(100), 
     "set the number of hash tables")
    //("and,A", 
    // po::value<uint>(&nand)->default_value(500), 
    // "set the number of hash functions per table")
    //("section,S", 
    // po::value<uint>(&nsec)->default_value(2), 
    // "set the number of sections per hash function")
    //("normalize,N", 
    // po::value<bool>(&normalize)->zero_tokens()->default_value(false),
    // "normalize the data onto unit hypersphere")
    ("random,R", 
     po::value<uint>(&seed)->default_value(123456789), 
     "set the seed for random generator")
   ;
}

LSH::
LSH(const LSH::Options& opts, bool forest)
  : ntab_(opts.ntab), nand_(opts.nand), nsec_(opts.nsec),  
    normalize_(opts.normalize), rng_(opts.seed)
{
  if (forest) return;

  assert(dblpow(nsec_,nand_) <= intpow(2,32));

  table_.resize(ntab_);
  for (uint i=0; i!=ntab_; ++i)
    table_[i] = HashTable(intpow(nsec_,nand_));
}

FeatureVector LSH::
get_feat(uint idx) const
{
  const HashData& d = get_data(idx);
  const uint begin = d.off_;
  const uint end = begin + d.win_;

  FeatureVector feat(N_FEAT*d.win_);  

  for (uint i=0, j=begin; j!=end; ++j)
    for (uint k=0; k!=N_FEAT; ++k)
      feat[i++] = mother_[j][k];
  
  return feat;
}

const HashData& LSH::
get_data(uint idx) const
{
  return data_[idx];
}

void LSH::
add_feat(const vector<FeatureVector>& f)
{
  mother_.insert(mother_.end(),f.begin(),f.end());
}

void LSH::
add_data(const HashData& d)
{
  data_.push_back(d);
}

void LSH::
dump() 
{
  print_function();
  print_index();
}

void LSH::
probe(vector<HashHit>& hit, const FeatureVector& x, 
      uint nprb, uint nthr)
{
  list<HashHit> tmp; // use list for massive push_back()

  for (uint i=0; i!=ntab_; ++i) {
    vector<HashValue> seq;
    multi_probe(seq,x,i,nprb);

    for (uint j=0; j!=seq.size(); ++j) {
      HashBacket::iterator k = table_[i][seq[j]].begin();
      HashBacket::iterator ke = table_[i][seq[j]].end();
      for ( ; k!=ke; ++k) {
	HashHit h;
	h.i = *k;
	h.s = 0.0;
	tmp.push_back(h); 
      }
    }
  }
  
  hit.assign(tmp.begin(),tmp.end());
  sort(hit.begin(),hit.end(),HashHitLess());
  vector<HashHit>::iterator he = unique(hit.begin(),hit.end(),HashHitSame());
  hit.erase(he,hit.end());
}

void LSH::
clear()
{
  // clear() in STL does not immediately free the memory 
  vector<FeatureVector>().swap(mother_);
  vector<HashData>().swap(data_);
  vector<HashTable>().swap(table_);
}

void LSH::
extend_heap(list<ProbeCandidate>& heap, list<ProbeCandidate>::iterator& hi, 
	    vector<double>& z, int size)
{
  ProbeCandidate c = *hi, c_shf, c_exp;
  c_shf = c_exp = c;

  int last = -1; 
  for (uint i=0; i!=size; ++i) {
    if (c.first[i]) last=i; 
  }
  if (last > size-2) last = size-2; // fail safe
  c_shf.second += z[last+1] - z[last];
  c_shf.first[last] = false;
  c_shf.first[last+1] = true;
  c_exp.second += z[last+1];
  c_exp.first[last+1] = true;
  heap.push_back(c_shf);
  heap.push_back(c_exp);

  heap.erase(hi);
}
