#ifndef __INC_LSH_H__
#define __INC_LSH_H__

#include <vector>
#include <list>
#include <deque>
#include <boost/program_options.hpp>

#include <limits.h>  // necessary for boost/serialization
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>

#include "Utility.h"

//---------- Definitions and Declaration of Hash Types ------------------

// sequence fragment 
struct HashData 
{
  HashData(uint sid, uint pos, uint win, uint off) 
    : sid_(sid), pos_(pos), win_(win), off_(off) {}
  HashData() {}
  ~HashData() {}

  uint sid_, pos_, win_, off_;

private:
  friend class boost::serialization::access;
  template <class Archive> 
  void serialize(Archive& ar, uint ver)
  {
    ar & sid_ & pos_ & win_ & off_;
  }
};

struct HashHit
{
  HashHit() {}
  ~HashHit() {}

  uint i;
  float s;
};

struct HashHitLess {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.i < y.i; }
};
struct HashHitLarger {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.i > y.i; }
};
struct HashHitSame {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.i == y.i; }
};

struct HashHitScoreLess {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.s < y.s; }
};
struct HashHitScoreLarger {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.s > y.s; }
};
struct HashHitScoreSame {
  bool operator()(const HashHit& x, const HashHit& y) const
  { return x.s == y.s; }
};

typedef unsigned long int HashValue;
typedef double WeightValue;
typedef std::vector<WeightValue> WeightVector;

//---------- Base Class for Various LSH Schemes -------------------------
//---------- Indyk and Motwani, 1998

class LSH
{
public:
  struct Options 
  {
    uint ntab;
    static const uint nand = 500; 
    static const uint nsec = 2;
    static const bool normalize = true;
    uint seed;

    void add_options(boost::program_options::options_description& desc);
  };

protected:
  typedef std::list<uint> HashBacket; 
  typedef std::vector<HashBacket> HashTable;
protected:
  typedef std::deque<bool> HashProbe; // do not use vector<bool> broken in STL
  typedef std::pair<HashProbe,double> ProbeCandidate;
  typedef std::pair<uint,double> HashMargin;

protected:
  LSH(const Options& opts, bool forest);
  LSH() : ntab_(0), nand_(0), nsec_(0), normalize_(false), rng_(0) {}
  ~LSH() {}

public:
  FeatureVector get_feat(uint idx) const;
  const HashData& get_data(uint idx) const;
  void add_feat(const std::vector<FeatureVector>& feat);
  void add_data(const HashData& d);
  void dump();
  virtual void make_function(uint nthr) = 0;
  virtual void make_index(uint nthr) = 0;
  virtual void print_function() = 0;
  virtual void print_index() = 0; 
  virtual void clear();
  virtual void probe(std::vector<HashHit>& hit, const FeatureVector& x, 
		     uint nprb, uint nthr);
  
protected:
  virtual HashValue hash_value(const FeatureVector& x, uint idx) = 0;

  // Lv et al, 2007;  Dong et al, 2008
  virtual void multi_probe(std::vector<HashValue>& seq, 
			   const FeatureVector& x, uint idx, uint nprb) = 0;

  void extend_heap(std::list<ProbeCandidate>& heap, 
		   std::list<ProbeCandidate>::iterator& hi,
		   std::vector<double>& z, int size);

  struct ProbeCandidateLess {
    bool operator()(const ProbeCandidate& x, const ProbeCandidate& y) const
    { return x.second < y.second; }
  };
  
  struct HashMarginLess {
    bool operator()(const HashMargin& x, const HashMargin& y) const
    { return x.second < y.second; }
  };

protected:
  std::vector<FeatureVector> mother_;
  std::vector<HashTable> table_;
  std::vector<HashData> data_;
  const uint ntab_;
  const uint nand_;
  const uint nsec_;
  const bool normalize_;
  RandomGenerator rng_;

private:
  friend class boost::serialization::access;
  template <class Archive> 
  void serialize(Archive& ar, const uint ver)
  {
    ar & mother_ & table_ & data_; 
    ar & const_cast<uint&>(ntab_) & const_cast<uint&>(nand_)
       & const_cast<uint&>(nsec_) & const_cast<bool&>(normalize_);
  }
};

#endif
