#include <iostream>
#include <cassert>
#include <cmath>

#include <boost/mpl/or.hpp> // necessary for syev()
#include <boost/numeric/bindings/lapack/syev.hpp>
#include <boost/numeric/bindings/lapack/syevd.hpp>
#include <boost/numeric/bindings/traits/std_vector.hpp>
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>

#include "Utility.h"

using namespace std;
using namespace boost;
namespace bnu = boost::numeric::ublas;
namespace bnbl = boost::numeric::bindings::lapack;

unsigned RandomShuffler::
operator()(unsigned i)
{
  uniform_int<> unif_dist(0,i-1);
  variate_generator<RandomGenerator&,uniform_int<> > unif_samp(rng_,unif_dist);

  return unif_samp();
}

unsigned long int 
intpow(const int a, const int b)
{
  return static_cast<unsigned long int>
    (pow(static_cast<double>(a), static_cast<double>(b)) );
}

double
dblpow(const int a, const int b)
{
  return static_cast<double>
    (pow(static_cast<double>(a), static_cast<double>(b)) );
}

FeatureValue
sqdist(const FeatureVector& x, const FeatureVector& y)
{
  assert(x.size() == y.size());

  FeatureValue sqd = 0.0;
  for(uint i=0; i!=x.size(); ++i) 
    sqd += pow(x[i]-y[i],2);

  return sqd;
}

FeatureValue
dist(const FeatureVector& x, const FeatureVector& y)
{
  return sqrt(sqdist(x,y));
}

FeatureVector
normalized_vector(const FeatureVector& x)
{
  FeatureVector xn(x);
  double norm = 0.0;

  for (uint i=0; i!=xn.size(); ++i) 
    norm += pow(xn[i],2);
  norm = sqrt(norm);
  for (uint i=0; i!=xn.size(); ++i) 
    xn[i] /= norm;

  return xn;
}

vector<double> 
random_normal_vector(const uint ndim, RandomGenerator& rng)
{
  vector<double> v(ndim);
  normal_distribution<double> norm_dist(0,1);
  variate_generator<RandomGenerator&,normal_distribution<double> > norm_samp(rng,norm_dist);

  for (uint i=0; i!=ndim; ++i) 
    v[i] = norm_samp();

  return v;
}

// generate a column-wise random rotation matrix 
// using Gram-Schmidt orthologonalization 
bnu::matrix<double> 
random_rotation_matrix(const uint ndim, RandomGenerator& rng)
{
  bnu::matrix<double> m(ndim,ndim);
  normal_distribution<double> norm_dist(0,1);
  variate_generator<RandomGenerator&,normal_distribution<double> > norm_samp(rng,norm_dist);

  for (uint i=0; i!=ndim; ++i) {
    bnu::matrix_column<bnu::matrix<double> > mc(m,i);
    bnu::vector<double> c(ndim);
    for (uint j=0; j!=ndim; ++j) 
      c(j) = norm_samp();
    for (uint j=0; j!=i; ++j) {
      bnu::matrix_column<bnu::matrix<double> > mt(m,j);
      c = c - bnu::inner_prod(c,mt) * mt / bnu::norm_2(mt);
    }
    mc = c / bnu::norm_2(c);
  } 

  return m;
}

// perform primary component analysis 
// using the eigenvalue decomposition algorithm in LAPACK
void 
pca(const vector<FeatureVector>& dat, 
    bnu::matrix<double>& pc, vector<double>& cr)
{
  const uint ndat = dat.size();
  const uint ndim = dat[0].size();

  for (uint i=0; i!=ndat; ++i) 
    assert(dat[i].size()==ndim);

  // calculate a covariance matrix
  bnu::symmetric_matrix<double> cov(ndim);
  vector<double> mean(ndim);
  for (uint i=0; i!=ndim; ++i) {
    double m = 0.0;
    for (uint j=0; j!=ndat; ++j)
      m += dat[j][i];
    mean[i] = m / ndat;
  }
  for (uint i=0; i!=ndim; ++i) {
    for (uint j=0; j<=i; ++j) {
      double c = 0.0;
      for (uint k=0; k!=ndat; ++k) 
	c += (dat[k][i]-mean[i]) * (dat[k][j]-mean[j]);
      cov(i,j) = c / ndat;
    }
  }

  bnu::matrix<double,bnu::column_major> cq(ndim,ndim);
  vector<double> cd(ndim);
  int info;

  for (uint i=0; i!=ndim; ++i) 
    for (uint j=0; j<=i; ++j) 
      cq(i,j) = cov(i,j);

  info = bnbl::syev('V','L',cq,cd,bnbl::optimal_workspace());
  //info = bnbl::syevd('V','L',cq,cd,bnbl::optimal_workspace()); // fast
  BOOST_UBLAS_CHECK(info==0,bnu::internal_logic());

  // we want primary directions (larger eigenvalues) first
  pc = cq;
  bnu::matrix<double> t(pc);
  for (uint i=0; i!=ndim; ++i) {
    bnu::matrix_column<bnu::matrix<double> > pcc(pc,i);
    bnu::matrix_column<bnu::matrix<double> > tc(t,ndim-i-1);
    pcc = tc;
  }

  cr = cd;
  reverse(cr.begin(),cr.end());
}

// calculate the square root of a synmetric matrix 
// using K = V^(T)*A*V -> K^(1/2) = V^(T)*A^(1/2)*V
vector<vector<double> >
sqrt_symmetric_matrix(const vector<vector<double> >& mat) 
{
  const uint ndim = mat.size();

  for (uint i=0; i!=ndim; ++i) 
    assert(mat[i].size()==ndim);
  for (uint i=0; i!=ndim; ++i) 
    for (uint j=0; j<i; ++j) 
      assert(mat[i][j]==mat[j][i]);

  vector<vector<double> > ret_mat(ndim,vector<double>(ndim));
  bnu::matrix<double> sqrt_mat(ndim,ndim); // K^(1/2)
  bnu::matrix<double> sqrt_val(ndim,ndim); // A^(1/2)
  bnu::matrix<double> vec(ndim,ndim); // V

  bnu::matrix<double,bnu::column_major> cq(ndim,ndim); 
  vector<double> cd(ndim); 
  int info;

  for (uint i=0; i!=ndim; ++i) 
    for (uint j=0; j<=i; ++j) 
      cq(i,j) = mat[i][j];

  info = bnbl::syev('V','L',cq,cd,bnbl::optimal_workspace());
  //info = bnbl::syevd('V','L',cq,cd,bnbl::optimal_workspace()); // fast
  BOOST_UBLAS_CHECK(info==0,bnu::internal_logic());

  for (uint i=0; i!=ndim; ++i) 
    if (cd[i]<0.0) cd[i]=0.0; 

  vec = cq;
  for (uint i=0; i!=ndim; ++i) 
    for (uint j=0; j!=ndim; ++j) 
      sqrt_val(i,j) = 0.0;
  for (uint i=0; i!=ndim; ++i) 
    sqrt_val(i,i) = sqrt(cd[i]);

  sqrt_mat = prod(bnu::matrix<double>(prod(vec,sqrt_val)),trans(vec));

  for (uint i=0; i!=ndim; ++i) {
    bnu::matrix_column<bnu::matrix<double> > col(sqrt_mat,i);
    copy(col.begin(),col.end(),ret_mat[i].begin());
  }

  return ret_mat;
}
